#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <unistd.h>

#define NUM_IN   2
#define NUM_HID1 8
#define NUM_HID2 8
#define NUM_OUT  1
#define MU       0.5
#define ALPHA    0.3

/* USAGE:
 *  init_nn();
 *  set in[] and target[]
 *  run_net() (the output will be in out[])
 *  backprop();
*/

float mu=MU;                         //learning rate
float mc=ALPHA;                         //momentum constant

float in[NUM_IN];                     //the two inputs (output values of nodes)
float hid1_i[NUM_HID1];                 //out values of the hidden nodes
float hid1[NUM_HID1];
#ifdef NUM_HID2
float hid2_i[NUM_HID2];
float hid2[NUM_HID2];
#endif
float out_i[NUM_OUT];
float out[NUM_OUT];
float target[NUM_OUT];

#ifdef NUM_HID2
float wa_in_hid1[NUM_IN][NUM_HID1];   //weights
float wa_hid1_hid2[NUM_HID1][NUM_HID2];
float wa_hid2_out[NUM_HID2][NUM_OUT];
float din[NUM_IN];
float dhid1[NUM_HID1];               //changes
float dhid2[NUM_HID2];
float dout[NUM_OUT];
float dwa_in_hid1[NUM_IN][NUM_HID1];   //weight changes
float dwa_hid1_hid2[NUM_HID1][NUM_HID2];
float dwa_hid2_out[NUM_HID2][NUM_OUT];
float dowa_in_hid1[NUM_IN][NUM_HID1];   //previous weight changes (for momentum)
float dowa_hid1_hid2[NUM_HID1][NUM_HID2];
float dowa_hid2_out[NUM_HID2][NUM_OUT];
#else
float wa_in_hid1[NUM_IN][NUM_HID1];     //weights
float wa_hid1_out[NUM_HID1][NUM_OUT];
float din[NUM_IN];
float dhid1[NUM_HID1];                  //changes
float dout[NUM_OUT];
float dwa_in_hid1[NUM_IN][NUM_HID1];    //weight changes
float dwa_hid1_out[NUM_HID1][NUM_OUT];
float dowa_in_hid1[NUM_IN][NUM_HID1];   //privious weight changes
float dowa_hid1_out[NUM_HID1][NUM_OUT];
#endif

pid_t getpid(void);
float sigmoid(float);
float normrand(void);
float floatrand(void);
void randweights(void);
void showweights(void);
void teach(int);
void back_prop(void);
void run_net(void);

void init_nn ( void ) {
  srand(getpid());
  randweights();
}

void showweights( void ) {
  int i,j;

  for (i=0;i<NUM_IN;i++) {
    for (j=0;j<NUM_HID1;j++) {
      printf ("%.2f ",wa_in_hid1[i][j]);
    }
  }
  printf("\n");
}

//back_prop: adjust the weights using the back propogation method
void back_prop( void ) {

  int i,j;
  float grad_out[NUM_OUT];
  #ifdef NUM_HID2
  float grad_hid2[NUM_HID2];
  #endif
  float grad_hid1[NUM_HID1];
  float tmp;

  #ifdef NUM_HID2
    //evaluate omega for all nodes besides input

    for (i=0;i<NUM_OUT;i++) { //output nodes
      grad_out[i]=out[i]*(1.0-out[i])*(target[i]-out[i]);
    }

    for (i=0;i<NUM_HID2;i++) { //second hidden layer
      tmp=0.0;
      for (j=0;j<NUM_OUT;j++) {
        tmp+=wa_hid2_out[i][j]*grad_out[j];
      }
      grad_hid2[i]=hid2[i]*(1.0-hid2[i])*tmp;
    }

    for (i=0;i<NUM_HID1;i++) { //first hidden layer
      tmp=0.0;
      for (j=0;j<NUM_HID2;j++) {
        tmp+=wa_hid1_hid2[i][j]*grad_hid2[j];
      }
      grad_hid1[i]=hid1[i]*(1.0-hid1[i])*tmp;
    }

    //now change the weights

    for (i=0;i<NUM_HID2;i++) {
      for (j=0;j<NUM_OUT;j++) {
        dwa_hid2_out[i][j]=(mu*grad_out[j]*hid2[i])+(dowa_hid2_out[i][j]*mc);
      }
    }

    for (i=0;i<NUM_HID1;i++) {
      for (j=0;j<NUM_HID2;j++) {
        dwa_hid1_hid2[i][j]=(mu*grad_hid2[j]*hid1[i])+(dowa_hid1_hid2[i][j]*mc);
      }
    }

    for (i=0;i<NUM_IN;i++) {
      for (j=0;j<NUM_HID1;j++) {
        dwa_in_hid1[i][j]=(mu*grad_hid1[j]*in[i])+(dowa_in_hid1[i][j]*mc);
      }
    }

    //change weights from deltas, and save to old weights

    for (i=0;i<NUM_HID2;i++) {
      for (j=0;j<NUM_OUT;j++) {
        wa_hid2_out[i][j]+=dwa_hid2_out[i][j];
        dowa_hid2_out[i][j]=dwa_hid2_out[i][j];
      }
    }

    for (i=0;i<NUM_HID1;i++) {
      for (j=0;j<NUM_HID2;j++) {
        wa_hid1_hid2[i][j]+=dwa_hid1_hid2[i][j];
        dowa_hid1_hid2[i][j]=dwa_hid1_hid2[i][j];
      }
    }

    for (i=0;i<NUM_IN;i++) {
      for (j=0;j<NUM_HID1;j++) {
        wa_in_hid1[i][j]+=dwa_in_hid1[i][j];
        dowa_in_hid1[i][j]=dwa_in_hid1[i][j];
      }
    }
  #else
    for (i=0;i<NUM_OUT;i++) { //output nodes
      grad_out[i]=out[i]*(1.0-out[i])*(target[i]-out[i]);
    }

    for (i=0;i<NUM_HID1;i++) { //hidden 
      tmp=0.0;
      for (j=0;j<NUM_OUT;j++) {
        tmp+=wa_hid1_out[i][j]*grad_out[j];
      }
      grad_hid1[i]=hid1[i]*(1.0-hid1[i])*tmp;
    }

    //now change the weights

    for (i=0;i<NUM_HID1;i++) {
      for (j=0;j<NUM_OUT;j++) {
        dwa_hid1_out[i][j]=(mu*grad_out[j]*hid1[i])+(dowa_hid1_out[i][j]*mc);
      }
    }

    for (i=0;i<NUM_IN;i++) {
      for (j=0;j<NUM_HID1;j++) {
        dwa_in_hid1[i][j]=(mu*grad_hid1[j]*in[i])+(dowa_in_hid1[i][j]*mc);
      }
    }
  
    //change weights from deltas, and save to old weights

    for (i=0;i<NUM_HID1;i++) {
      for (j=0;j<NUM_OUT;j++) {
        wa_hid1_out[i][j]+=dwa_hid1_out[i][j];
        dowa_hid1_out[i][j]=dwa_hid1_out[i][j];
      }
    }

    for (i=0;i<NUM_IN;i++) {
      for (j=0;j<NUM_HID1;j++) {
        wa_in_hid1[i][j]+=dwa_in_hid1[i][j];
        dowa_in_hid1[i][j]=dwa_in_hid1[i][j];
      }
    }
  #endif
}

//run_net: hopefully this is self explanatory...
void run_net ( void ) { 
  int i,j;
  float tmp;

  //the array in[] needs to have been set manually before calling this function

  for (i=0;i<NUM_HID1;i++) {
    tmp=0.0;
    for (j=0;j<NUM_IN;j++) {
      tmp+=in[j]*wa_in_hid1[j][i];
    }
    hid1_i[i]=tmp;
    hid1[i]=sigmoid(hid1_i[i]);
  }

  #ifdef NUM_HID2
    for (i=0;i<NUM_HID2;i++) {
      tmp=0.0;
      for (j=0;j<NUM_HID1;j++) {
        tmp+=hid1[j]*wa_hid1_hid2[j][i];
      }
      hid2_i[i]=tmp;
      hid2[i]=sigmoid(hid2_i[i]);
    }

    for (i=0;i<NUM_OUT;i++) {
      tmp=0.0;
      for (j=0;j<NUM_HID2;j++) {
        tmp+=hid2[j]*wa_hid2_out[j][i];
      } 
      out_i[i]=tmp;
      out[i]=sigmoid(out_i[i]);
    }
  #else
    for (i=0;i<NUM_OUT;i++) {
      tmp=0.0;
      for (j=0;j<NUM_HID1;j++) {
        tmp+=hid1[j]*wa_hid1_out[j][i];
      }
      out_i[i]=tmp;
      out[i]=sigmoid(out_i[i]);
    }
  #endif

  //the output has been stored in out[]
}

//randweights: set the weights for the first run
void randweights ( void ) {
  int i,j;

  srand(getpid());

  #ifdef NUM_HID2
   for (i=0;i<NUM_IN;i++) {
     for (j=0;j<NUM_HID1;j++) {
       wa_in_hid1[i][j]=normrand();
     }
   }

   for (i=0;i<NUM_HID1;i++) {
     for (j=0;j<NUM_HID2;j++) {
       wa_hid1_hid2[i][j]=normrand();
     }
   }

   for (i=0;i<NUM_HID2;i++) {
     for (j=0;j<NUM_OUT;j++) {
       wa_hid2_out[i][j]=normrand();
     }
   }
  #else
   for (i=0;i<NUM_IN;i++) {
     for (j=0;j<NUM_HID1;j++) {
       wa_in_hid1[i][j]=normrand();
     }
   }

   for (i=0;i<NUM_HID1;i++) {
     for (j=0;j<NUM_OUT;j++) {
       wa_hid1_out[i][j]=normrand();
     }
   }
  #endif
}

//sigmoid: calculate the sigmoid function value
//regulates between 0 and 1 on an S curve
//in: the value to be calculated
float sigmoid ( float x ) {
  return (1/(1+exp(-x)));
}

//normrand: return a random number on the 
//interval [-1,1]
float normrand( void ) {
  return ((2.0*(((float)(rand()))/((float)(RAND_MAX))))-1.0);
}

//floatrand: 0-->1
float floatrand( void ) {
  float r=normrand();

  if (r>0.0) {
    return r;
  } else {
    return -r;
  }
}

